/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * License); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * AS IS BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

/*
 * Copyright (c) 2019, Open AI Lab
 * Author: bingzhang@openailab.com
 */
#include "operator/expanddims.hpp"
#include "static_graph.hpp"

namespace TEngine {

bool ExpandDims::InferShape(const std::vector<TEngine::TShape>& ishape, std::vector<TEngine::TShape>& oshape, int layout)
{
    const TShape& input = ishape[0];

    const std::vector<int>& in_dim = input.GetDim();
    std::vector<int> out_dim;
    int in_size = in_dim.size();
    for(int i = 0; i < in_size; i++){
        out_dim.push_back(in_dim[i]);
    }
    out_dim.push_back(0);

    int out_size = in_size + 1;
    for(int i = 0; i < out_size; i++){
        if( i < param_.axis){
            out_dim = in_dim;
        } else if(i == param_.axis){
            out_dim[i] = 1;
        } else {
            out_dim[i] = in_dim[i-1];
        }
    }

    TShape shape;
    shape.SetDim(out_dim);
    shape.SetDataLayout(input.GetDataLayout());
    oshape[0] = shape;   
    
    return true;
}

void ExpandDims::SetSchema(void)
{
    Input({"input:float32"})
        .Output({"output:float32"})
        .SetAttr("axis", 0)
        .SetDoc(R"DOC(ExpandDims Operator)DOC");
}

}    // namespace TEngine
